/**
*Algoritmo Gentico para tratar el CVRP
*@author Fernando Michel Tavera
*@version 0.6
*/
public class EjecutaAG{

/**
*Permite la ejecucin del programa
*
*/
	public static void main(String[] args){
		double[] pesos=new double[3];
		int[] tiempos = new int[3];
		int[] numVs = new int[3];
		
		String ruta="./ejemplares/";
		String nombreArch =ruta+"10.vrp";
		int maxV=15;
		int generaciones=1000000;
		int penalizacion=50;
		try{
			if (args.length>3)
				generaciones=Integer.parseInt(args[3]);
			if (args.length>2)
				penalizacion=Integer.parseInt(args[2]);
			if (args.length>1)
				maxV=Integer.parseInt(args[1]);
			if (args.length>0)
				nombreArch=args[0];
		}catch(NumberFormatException e){
			e.printStackTrace();
			System.exit(0);
		}
		
		VRP ejemplar = LeeArchivo.leer(nombreArch, maxV, penalizacion);
		int numV=0;
		double total = 0;
		Genetico ag;
		GeneticoAnidado aga;
		Individuo resultado;
		Individuo mejor;
		double costoMejor;
		int tAntes;
		int tDespues;
		String texto;
		for (int i = 0; i<14;++i){
			nombreArch = ruta+(i+1)+".vrp";
			System.out.println("\n"+(i+1)+" normal:");
			maxV=6;
			if (i==2 || i==5 || i== 10 || i== 11) maxV=12;
			if (i==1 || i==3 || i== 7 || i== 12 || i== 13) maxV=18;
			if (i==4 || i== 6 ) maxV=24;
			if (i==8) maxV=29;
			if (i== 9) maxV=35;
			ejemplar = LeeArchivo.leer(nombreArch, maxV, penalizacion);
			ag = new Genetico(ejemplar);
			tAntes = (int)(System.currentTimeMillis()/1000);
			resultado = ag.corrida(generaciones);
			tDespues = (int)(System.currentTimeMillis()/1000);
			total = (resultado.aptitud()==0? -1 : ag.pesoMax() - resultado.aptitud());
			numV = (total<0?-1 : resultado.numRutasUsadas());
			pesos[0] =costoMejor = total;
			mejor=new Individuo(ejemplar.numClientes());
			mejor.copia(resultado);
			numVs[0] = numV;
			tiempos[0] = 1+(tDespues - tAntes);
System.out.print(", tiempo: "+tiempos[0]+" segundos\n");
			System.out.println("\n"+(i+1)+" anidado:");
			aga = new GeneticoAnidado(ejemplar);
			tAntes = (int)(System.currentTimeMillis()/1000);
			resultado = aga.corrida(generaciones, false);
			tDespues = (int)(System.currentTimeMillis()/1000);
			total = (resultado.aptitud()==0?-1 : aga.pesoMax() - resultado.aptitud());
			numV = (total<0?-1 : resultado.numRutasUsadas());
			pesos[1] = total;
			if (costoMejor<0 ||(total<costoMejor && total>0)){
				costoMejor = total;
				mejor.copia(resultado);
			}
			numVs[1] = numV;
			tiempos[1] = 1+(tDespues - tAntes);
System.out.print(", tiempo: "+tiempos[1]+" segundos\n");
			System.out.println("\n"+(i+1)+" refinado:");
			tAntes = (int)(System.currentTimeMillis()/1000);
			resultado = aga.refinar(mejor);
			tDespues = (int)(System.currentTimeMillis()/1000);
			total = (resultado.aptitud()==0?-1 : aga.pesoMax() - resultado.aptitud());
			numV = (total<0?-1 : resultado.numRutasUsadas());
			pesos[2] = total;
			if (costoMejor<0 ||(total<costoMejor && total>0)){
				costoMejor = total;
				mejor.copia(resultado);
			}
			numVs[2] = numV;
			tiempos[2] = 1+(tDespues - tAntes);
System.out.print(", tiempo: "+tiempos[2]+" segundos\n");
			texto = "\n\n\t\tnormal\t\t\tanidado\t\t\tanidado2\nid\tcosto\tnv\ttiempo\tcosto\tnv\ttiempo\tcosto\tnv\ttiempo\n";
			texto = texto+(i+1)+"\t"+pesos[0]+"\t"+numVs[0]+"\t"+tiempos[0]+"\t"+pesos[1]+"\t"+numVs[1]+"\t"+tiempos[1]+"\t"+pesos[2]+"\t"+numVs[2]+"\t"+tiempos[2]+"\n\n";
			texto = texto+"mejor ruta:\n"+imprimeInd(mejor, ejemplar.numVehiculos());
			nombreArch = ruta+(i+1);
			if (penalizacion>0)
				nombreArch = nombreArch +"p";
			nombreArch = nombreArch +".txt";
			EscribeArchivo.escribe(nombreArch, texto);
			System.out.println(texto+"\n");
		}
	}

/*
*Imprime la solucin representada por el individuo
*
*/
	private static String imprimeInd(Individuo ind, int numV){
		int numC = ind.tamCromo();
		int[][] rutas = new int[numV][numC+1];
		int r, n, c;
		for (int i = 0; i < numC; ++i){
			c = ind.orden(i);
			r = ind.ruta(c);
			n = ++rutas[r][0];
			rutas[r][n] = c+1;
		}
		n = 1;
		String s="";
		for (int i = 0; i < numV; ++i){
			if (rutas[i][0] > 0){
				s=s+"\nRuta para el vehiculo "+(n++)+":\n\tc0";
				for (int j = 1; j<= rutas[i][0]; ++j){
					s=s+" -> "+LeeArchivo.coordenadas(rutas[i][j]);
				}
				s=s+" -> c0\n";
			}
		}
		return s;
	}
}