import java.util.Random;
/**
*Algoritmo Gentico para tratar el CVRP
*@author Fernando Michel Tavera
*@version 0.8
*/
public class Genetico{
	protected int mitadP;
	protected Individuo[] poblacion;
	protected float probCruce;
	protected float probMutacion;
	protected float probMutaGen;
	protected Individuo mejorInd;
	protected Random rand;
	protected double maxApt;
	protected Individuo[] nuevaPob;
	protected int elitismo;
	protected Individuo[] elite;
	
	protected VRP ejemplar;
	
	//arreglos usados para el proceso de evaluacion
	protected float[] capV;
	protected int[] distV;
	protected int[] ultimo;

/**
*Inicializa el AG con el ejemplar del CVRP
*
*/
	public Genetico(VRP ejem){
		ejemplar = ejem;
		mitadP = 10;
		probCruce = 0.985f;
		probMutacion = 0.1f;
		probMutaGen = 0.35f;
		elitismo = 5;
		rand = new Random();
		generaInicial();
		maxApt = calculaMax();
	}
/**
*Inicializa el algoritmo con el ejemplar del CVRP y los parametros de la heurstica
*
*/
	public Genetico(VRP ejem, int s, float pc, float pm, float pm2, int numElite){
		ejemplar = ejem;
		mitadP = s;
		probCruce = pc;
		probMutacion = pm;
		probMutaGen = pm2;
		elitismo = numElite;
		rand = new Random();
		generaInicial();
		maxApt = calculaMax();
	}
/**
*Ejecuta el AG hasta alcanzar un nmero de generaciones
*
*/
	public Individuo corrida(int numGeneraciones){
		/**/
		int avance = numGeneraciones/10;
		char[] charAvance = new char[]{'9','8','7','6','5','4','3','2','1','0'};
		int indAvance=0;
		int cntAvance=0;/**/
		
		double tamRuleta;
		double adap;
		int padre1, padre2, puntoCruce;
		for (int generacion = 0; generacion < numGeneraciones; ++generacion){
			/**/
			if (cntAvance++>=avance){
				System.out.print(charAvance[indAvance++]);
				cntAvance=0;
			}
			/**/
			tamRuleta = 0;
			for (int i = 0; i < poblacion.length; ++i){
				adap = evalua(poblacion[i]);
				pruebaElite(poblacion[i], adap);
				tamRuleta += adap;
			}
			for (int i = 0; i < mitadP; i++){
				padre1 = seleccion(tamRuleta);
				padre2 = seleccion(tamRuleta);
				puntoCruce = rand.nextInt(ejemplar.numClientes());
				if (rand.nextFloat() < probCruce){
					cruza(2*i, padre1, padre2, puntoCruce);
				}else{
					nuevaPob[2*i].copia(poblacion[padre1]);
					nuevaPob[2*i +1].copia(poblacion[padre2]);
				}
				if (rand.nextFloat() < probMutacion)
					muta(2*i);
				if (rand.nextFloat() < probMutacion)
					muta(2*i+1);
			}
			for (int i = 0; i < elitismo; ++i){
				poblacion[i].copia(elite[i]);
			}
			for (int i = elitismo; i < poblacion.length; ++i){
				poblacion[i].copia(nuevaPob[i-elitismo]);
			}
		}
System.out.print("\nvalor final: "+(maxApt -elite[0].aptitud())+"(aptitud "+elite[0].aptitud()+")");
		return elite[0];
	}
/**
*Evala la aptitud de un individuo, actualiza el registro del mismo y regresa el valor de aptitud calculado
*
*/
	public void pruebaElite(Individuo ind, double adap){
		double adapElite;
		for (int j=elitismo-1;j>=0;--j){
			 adapElite= evalua(elite[j]);
			 if (adap > adapElite){
				 if(j<elitismo-1)
					 elite[j+1].copia(elite[j]);
				 elite[j].copia(ind);
				/**if (j==0)
				System.out.print("*");/**/
			}
		}
	}
/**
*Evala la aptitud de un individuo, actualiza el registro del mismo y regresa el valor de aptitud calculado
*
*/
	public double evalua(Individuo ind){
/*	for (int i=0;i<ejemplar.numClientes();++i){
		System.out.println(i+","+demanda[i]);
	}*/
		
		if (ind.evaluado()){
			double apt = ind.aptitud();
			return apt;
		}
		int nC = ejemplar.numClientes();
		int nV = ejemplar.numVehiculos();
		float cap = ejemplar.capacidad();
		double dMax = ejemplar.distanciaMaxima();
		int cliente;
		int vehiculo;
		double total = 0.0d;
		//limpia los arreglos auxiliares
		for (int i = 0; i < nV; ++i){
			capV[i] = 0.0f;
			distV[i] = 0;
			ultimo[i] = 0;
		}
		for (int i = 0; i < nC; ++i){
			cliente = ind.orden(i);
			vehiculo = ind.ruta(cliente);
			capV[vehiculo] += ejemplar.demanda(cliente);
			distV[vehiculo] += ejemplar.distancia(ultimo[vehiculo],cliente+1);
			//procedimiento de correccion
			if (capV[vehiculo] > cap || distV[vehiculo] > dMax){
				capV[vehiculo] -= ejemplar.demanda(cliente);
				distV[vehiculo] -= ejemplar.distancia(ultimo[vehiculo],cliente+1);
				vehiculo=-1;
				for (int j=0;j<nV && vehiculo<0;++j){
					if (capV[j]+ejemplar.demanda(cliente)<=cap && distV[j]+ ejemplar.distancia(ultimo[j],cliente+1) <=dMax){
						vehiculo=j;
					}
				}
				if (vehiculo<0){
					ind.marcar(0);
					return 0;
				}else{
					ind.cambiaRuta(cliente, vehiculo);
					capV[vehiculo]+=ejemplar.demanda(cliente);
					distV[vehiculo] += ejemplar.distancia(ultimo[vehiculo],cliente+1);
				}	
			}
			total += ejemplar.distancia(ultimo[vehiculo],cliente+1);
			ultimo[vehiculo] = cliente+1;
		}
		//costo de regresar al deposito
		for (int i = 0; i < nV; ++i){
			total += ejemplar.distancia(ultimo[i],0);
		}
		double adap = maxApt - total;
		if (ejemplar.penalizacion()>0){
			adap -= ejemplar.penalizacion() * ind.numRutasUsadas();
		}
		ind.marcar(adap);
		return adap;
	}
/**
*Selecciona un individuo de la poblacin con el algoritmo de ruleta
*
*/
	protected int seleccion(double tamRuleta){
		if (tamRuleta <=0){
			return (rand.nextInt(poblacion.length));
		}
		double punto = rand.nextDouble()*tamRuleta;
		int ret = -1;
		int i = 0;
		while (ret < 0 && i < poblacion.length){
			if (punto < poblacion[i].aptitud()){
				ret = i;
			}else{
				punto -= poblacion[i].aptitud();
				++i;
			}
		}
		return ret;
	}
/**
*Lleva a cabo el cruzamiento entre dos individuos
*
*/
	protected void cruza(int pos, int p1, int p2, int puntoCruce){
		nuevaPob[pos].cruzaRutas(poblacion[p1], poblacion[p2], puntoCruce);
		nuevaPob[pos+1].cruzaRutas(poblacion[p2], poblacion[p1], puntoCruce);
		nuevaPob[pos].cruzaOrden(poblacion[p1], poblacion[p2], puntoCruce);
		nuevaPob[pos+1].cruzaOrden(poblacion[p2], poblacion[p1], puntoCruce);
	}
/**
*Lleva a cabo la mutancion de un individuo
*
*/
	protected void muta(int pos){
		nuevaPob[pos].mutaRutas(probMutaGen, rand);
		nuevaPob[pos].mutaOrden(probMutaGen, rand);
	}
/**
*Regresa el valor de aptitud mximo
*
*/
	public double pesoMax(){
		return maxApt;
	}

/*
*auxiliar que construye la poblacin inicial
*/
	protected void generaInicial(){
		int n = ejemplar.numClientes();
		int m = ejemplar.numVehiculos();
		int tam = (2*mitadP);
		poblacion = new Individuo[tam+elitismo];
		nuevaPob = new Individuo[tam];
		for (int i = 0; i< tam ; ++i){
			poblacion[i] = new Individuo(n, m, rand);
			nuevaPob[i] = new Individuo(n);
		}
		elite = new Individuo[elitismo];
		for (int i = 0; i< elitismo ; ++i){
			poblacion[tam + i] = new Individuo(n, m, rand);
			elite[i] = new Individuo(n);
			elite[i].copia(poblacion[0]);
		}
		capV = new float[m];
		distV = new int[m];
		ultimo = new int[m];
	}

/*
*auxiliar que calcula una cota superior al valor de aptitud de cualquier Individuo
*/
	protected double calculaMax(){
		int numC = ejemplar.numClientes();
		double max = 0;
		float val;
		for (int i = 0; i <= numC; ++i){
			for (int j = 0; j <= numC; ++j){
				val = ejemplar.distancia(i,j);
				if (val > max)
					max = ejemplar.distancia(i,j);
			}
		}
		max *= 2*numC;
		val = ejemplar.penalizacion();
		if (val > 0)
			max += val * ejemplar.numVehiculos();
		return max;
	}
}