import java.util.Random;
/**
*Algoritmo Gentico para tratar el CVRP
*@author Fernando Michel Tavera
*@version 2.0
*/
public class GeneticoAnidado extends Genetico{
	
	protected int[][] datosRuta;
	protected int[] ordenTmp;
	protected boolean exacto=false;//si es true, se obtiene el optimo exacto para el cromosoma de orden
	protected boolean elite2;
	private Individuo superElite, tmpElite;
	

/**
*Inicializa el AG con los datos del ejemplar
*
*/
	public GeneticoAnidado(VRP ejemplar){
		super(ejemplar);
		exacto = elite2 = false;
		generaDatosRuta();
	}
/**
*Inicializa el algoritmo con el ejemplar del CVRP y los parametros de la heurstica
*
*/
	public GeneticoAnidado(VRP ejem, int s, float pc, float pm, float pm2, int numElite){
		super(ejem, s, pc, pm, pm2, numElite);
		exacto = elite2 = false;
		generaDatosRuta();
	}

	public Individuo corrida(int generaciones, boolean e2){
		elite2 = e2;
		if (elite2){
			superElite = new Individuo(ejemplar.numClientes());
			tmpElite = new Individuo(ejemplar.numClientes());
			superElite.copia(elite[0]);
			pruebaElite2(superElite);
		}
		Individuo sol = corrida(generaciones);
		if (elite2){
			return superElite;
		}else{
			//boolean ex=exacto;
			//exacto=true;
			//evalua(sol);
			//exacto = ex;
			return sol;
		}
		
	}

	public Individuo refinar(Individuo ind){
		Individuo ref=new Individuo(ejemplar.numClientes());
		ref.copia(ind);
		if (superElite == null){
			superElite = new Individuo(ejemplar.numClientes());
			tmpElite = new Individuo(ejemplar.numClientes());
			superElite.marcar(0);
		}
		boolean ex=exacto;
		exacto=true;
		evalua(ref);
		exacto = ex;
		return ref;
	}
/*
*
*/
	protected void generaDatosRuta(){
		datosRuta = new int[ejemplar.numVehiculos()][ejemplar.numClientes()+1];
		ordenTmp = new int[ejemplar.numClientes()];
	}
/**
*
*/
@Override
	public void pruebaElite(Individuo ind, double adap){
		double adapElite;
		for (int j=elitismo-1;j>=0;--j){
			 adapElite= evalua(elite[j]);
			 if (adap > adapElite){
				 if(j<elitismo-1)
					 elite[j+1].copia(elite[j]);
				 if(elite2 && j==0){
					pruebaElite2(ind);
				 }
				 elite[j].copia(ind);
				/**if (j==0)
				System.out.print("*");/**/
			}
		}
	}
/**
*
*/
	protected void pruebaElite2(Individuo ind){
		boolean ex=exacto;
		tmpElite.copia(ind);
		exacto=true;
		double val=evalua(tmpElite);
		if (val>superElite.aptitud()){
			superElite.copia(tmpElite);
		}
		exacto = ex;
	}
/*
*
*/
	public double evalua(Individuo ind){
		if ((!exacto) && ind.evaluado()){
			double apt = ind.aptitud();
			return apt;
		}
		super.evalua(ind);
		int tam = ejemplar.numClientes();
		int j=0;
		for (int i=0;i<datosRuta.length;++i){
			datosRuta[i][0]=0;
		}
		for (int i =0;i < tam; ++i){
			j=ind.ruta(i);
			datosRuta[j][0]++;
			datosRuta[j][datosRuta[j][0]]=i+1;
		}
		j=0;
		for (int i =0;i < datosRuta.length; ++i){
			if(datosRuta[i][0]>0){
				procesaRuta(datosRuta[i],j, exacto);
				j+=datosRuta[i][0];
			}
		}
		ind.cambiaOrden(ordenTmp);
		return super.evalua(ind);
	}
	/*
	*
	*/
		protected void procesaRuta(int[] ruta, int offset, boolean bnb){
			double cota =0;
			int tam = ruta[0];
			int[] ordenRuta = new int[tam];
			boolean[] revisados = new boolean[tam];
			float min;
			float dist;
			int anterior=0;
			int indMin=0;
			float minDist=-1;
			for (int i=0; i< tam; ++i){
				min = -1;
				for (int j=0; j<tam;++j){
					if (!revisados[j]){
						dist = ejemplar.distancia(anterior, ruta[j+1]);
						if(bnb && (dist<minDist || minDist<0))
							minDist = dist;
						if (dist<min || min<0){
							indMin=j+1;
							min=dist;
						}
					}
				}
				cota+=min;
				anterior = ruta[indMin];
				ordenRuta[i] = anterior-1;
				revisados[indMin-1]=true;
			}
			cota+=ejemplar.distancia(ordenRuta[tam-1]+1, 0);
			if (bnb){
				branchAndBound(ordenRuta, cota, minDist);
			}
			System.arraycopy(ordenRuta, 0, ordenTmp, offset, tam);
		}

	/*
	*
	*/
		protected void branchAndBound(int[] ordenRuta, double cota, float minDist){
			int tam = ordenRuta.length;
			int[] ordenNuevo = new int[tam];
			System.arraycopy(ordenRuta, 0, ordenNuevo, 0, tam);
			int[] indiceRuta = new int[tam];
			boolean[] visitado = new boolean[tam];
			float[] pesoParcial = new float[tam+1];
			int nivel=0;
			int indActual=-1;
			int clienteActual=0, clienteAnterior=0;
			while(nivel>=0){
				if(++indActual>=tam){
					visitado[indiceRuta[nivel]]=false;
					if(--nivel>=0){
						indActual=indiceRuta[nivel];
						visitado[indActual]=false;
						clienteAnterior=nivel==0?0:1+ordenRuta[indiceRuta[nivel-1]];
					}
				}else{
					if(!visitado[indActual]){
						indiceRuta[nivel]=indActual;
						clienteActual=ordenRuta[indActual]+1;
						pesoParcial[nivel]=ejemplar.distancia(clienteAnterior,clienteActual);
						if (nivel > 0)
							pesoParcial[nivel]+=pesoParcial[nivel-1];
						if (pesoParcial[nivel]+(minDist*(tam-nivel))<cota){
							if(++nivel>=tam){
								pesoParcial[tam]=pesoParcial[tam-1]+ejemplar.distancia(clienteActual,0);

								if (pesoParcial[tam]<cota){
									//nuevo orden optimo encontrado
									cota = pesoParcial[tam-1];
									for(int i=0;i<tam;++i){
										ordenNuevo[i]=ordenRuta[indiceRuta[i]];
									}
								}
								--nivel;
							}else{
								visitado[indActual]=true;
								clienteAnterior=ordenRuta[indActual]+1;
								indActual = -1;
							}
						}else{
							visitado[indActual]=false;
							}
					}
				}
				
			}
			System.arraycopy(ordenNuevo, 0, ordenRuta, 0, tam);
		}
}