import java.util.Random;
/**
*Individuo que representa una solucin al CVRP
*@author Fernando Michel Tavera
*@version 2.0
**/
public class Individuo{
	private int tamCromo;
	private int numRutas;
	private int[] ruta;
	private int[] orden;
	private double aptitud;
	private boolean evaluado;

/**
*Construye un individuo aleatoriamente
*
**/
	public Individuo(int tamC, int numR, Random rand){
		tamCromo = tamC;
		numRutas = numR;
		evaluado = false;
		aptitud = 0;
		ruta = new int[tamCromo];
		orden = new int[tamCromo];
		for (int i = 0; i < tamCromo; ++i){
			ruta[i] = rand.nextInt(numRutas);
			orden[i] = i;
		}
		int j = 0;
		int tmp;
		for (int i = 0; i < tamCromo; ++i){
			j = rand.nextInt(tamCromo);
			tmp = orden[i];
			orden[i] = orden[j];
			orden[j] = tmp;
		}
	}
/**
*Construye un individuo vacio
*
**/
	public Individuo(int tamC){
		tamCromo = tamC;
		numRutas = tamC;
		evaluado = false;
		aptitud = 0;
		ruta = new int[tamCromo];
		orden = new int[tamCromo];
	}
/**
*Construye una copia de un Individuo
*
**/
	public void copia(Individuo ind){
		evaluado = ind.evaluado();
		aptitud = ind.aptitud();
		numRutas = ind.numRutas();
		for (int i = 0; i < tamCromo; ++i){
			ruta[i] = ind.ruta(i);
			orden[i] = ind.orden(i);
		}
	}

/**
*
*
**/
	public void cruzaRutas(Individuo p1, Individuo p2, int pCruce){
		evaluado = false;
		aptitud = 0;
		numRutas = p1.numRutas();
		boolean[] aparece = new boolean[tamCromo];
		for (int i = 0; i < pCruce; ++i){
			ruta[i] = p1.ruta(i);
		}
		for (int i = pCruce; i < tamCromo; ++i){
			ruta[i] = p2.ruta(i);
		}
	}
/**
*
*
**/
	public void cruzaOrden(Individuo p1, Individuo p2, int pCruce){
		boolean[] aparece = new boolean[tamCromo];
		for (int i = 0; i < pCruce; ++i){
			orden[i] = p1.orden(i);
			aparece[orden[i]] = true;
		}
		int j = pCruce;
		for (int i = 0; i < tamCromo; ++i){
			if (j<tamCromo && !aparece[p2.orden(i)]){
				orden[j] = p2.orden(i);
				j++;
			}
		}
	}

/**
*Devuelve el tamao del cromosoma
*
**/
	public int tamCromo(){
		return tamCromo;
	}
/**
*Devuelve la ruta asignada al cliente i
*
**/
	public int ruta(int i){
		return ruta[i];
	}
/**
*Cambia la ruta asignada al cliente i por r
*
**/
	public void cambiaRuta(int i, int r){
		ruta[i]=r;
	}
/**
*Devuelve el ndice del i-simo cliente visitado
*
**/
	public int orden(int i){
		return orden[i];
	}
/**
*Redefine el orden de forma laeatoria
*
**/
	public void cambiaOrden(int[] nuevoOrden){
		evaluado = false;
		aptitud = 0;
		for (int i = 0; i < tamCromo; ++i){
			orden[i] = nuevoOrden[i];
		}
	}
/**
*Devuelve el valor de aptitud del Individuo, o 0 si el Individuo no ha sido evaluado
*
**/
	public double aptitud(){
		return aptitud;
	}
/**
*Devuelve true si el Individuo ya ha sido evaluado, false en otro caso
*
**/
	public boolean evaluado(){
		return evaluado;
	}
/**
*Devuelve el nmero de rutas del Individuo
*
**/
	public int numRutas(){
		return numRutas;
	}
/**
*Devuelve el nmero de rutas usadas en la solucin representada por el Individuo
*
**/
	public int numRutasUsadas(){
		boolean[] usada = new boolean[numRutas];
		for (int i = 0; i < tamCromo; ++i){
			usada[ruta[i]] = true;
		}
		int num = 0;
		for (int i = 0; i < numRutas; ++i){
			if (usada[i])
				++num;
		}
		return num;
	}
/**
*lleva a cabo la mutacin de este Individuo
*
**/
	public void mutaRutas(float prob, Random rand){
		int j = 0;
		for (int i = 0; i < tamCromo; ++i){
			if (rand.nextFloat() < prob )
				ruta[i] = rand.nextInt(numRutas);
		}
	}
/**
*lleva a cabo la mutacin de este Individuo
*
**/
	public void mutaOrden(float prob, Random rand){
		int j = 0;
		int tmp;
		for (int i = 0; i < tamCromo; ++i){
			if (rand.nextFloat() < prob ){
				j = rand.nextInt(tamCromo);
				tmp = orden[i];
				orden[i] = orden[j];
				orden[j] = tmp;
			}
		}
	}
/**
*marca al Individuo como evaluado y asigna el valor de aptitud
**/
	public void marcar(double apt){
		aptitud = apt;
		evaluado = true;
	}
}